package com.gct.aoid.algorithm;

import com.alibaba.fastjson.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MarkerFactory;

import java.util.Arrays;
import java.util.List;

import static java.lang.Math.log;
import static java.lang.Math.pow;

/**
 * dec:
 * <p>
 * PumpCardAlgo
 * </p>
 *
 * @author lloyd
 * @version 1.0.1
 * @since 2022/9/27 上午10:16
 */
public class PumpCardAlgo {

    /**
     * 参数个数
     */
    private static final int p_num = 12;

    /**
     * 傅立叶系数矩阵
     */
    private Double[][] p_matrix;

    /**
     * 计算阻尼数据
     *
     * @param A_r   杆柱横截面积 平方米
     * @param miu_o 原油粘度 pa /s
     * @param D_t   油管直径 m
     * @param D_r   杆直径 m
     * @param rou_r 杆柱密度 kg/立方米
     * @param w
     * @param x     深度 m
     * @param a
     * @return
     */
    private double c_C(double A_r, double miu_o, double D_t, double D_r, double rou_r, double w, double x, double a) {
        double m = D_t / D_r;
        double B1 = (m * m - 1) / (2 * log(m)) - 1;
        double B2 = pow(m, 4) - 1 - pow(m * m - 1, 2) / log(m);
        double c = 2 * Math.PI * miu_o / rou_r / A_r *
                (1 / log(m) + 4 / B2 * (B1 + 1) * (B1 + 2 / (w * x / a / Math.sin(w * x / a) + Math.cos(w * x / a))));
        return c;
    }


    void init_matrix(int n) {
        p_matrix = new Double[p_num][];
        for (int i = 0; i < p_num; ++i) {
            p_matrix[i] = new Double[n];
            for (int j = 0; j < n; ++j) {
                p_matrix[i][j] = 0d;
            }
        }
    }

    private static final int alpha_n = 0;
    private static final int beta_n = 1;
    private static final int sigma_n = 2;
    private static final int tau_n = 3;
    private static final int nu_n = 4;
    private static final int delta_n = 5;
    private static final int kappa_n = 6;
    private static final int mu_n = 7;
    private static final int O_n = 8;
    private static final int P_n = 9;
    private static final int O_n_2 = 10;
    private static final int P_n_2 = 11;


    private double ch(double x) {
        return (Math.exp(x) + Math.exp(-x)) * 0.5;
    }

    private double sh(double x) {
        return (Math.exp(x) - Math.exp(-x)) * 0.5;
    }


    private void c_p_matrix_common(double E_r, double A_r, double x, int p_num_cur, int n_cur) {
        switch (p_num_cur) {
            case kappa_n: {
                p_matrix[p_num_cur][n_cur] = (p_matrix[sigma_n][n_cur] * p_matrix[alpha_n][n_cur] +
                        p_matrix[sigma_n][n_cur] * p_matrix[beta_n][n_cur]) /
                        E_r / A_r /
                        (pow(p_matrix[alpha_n][n_cur], 2) + pow(p_matrix[beta_n][n_cur], 2));
                break;
            }
            case mu_n: {
                p_matrix[p_num_cur][n_cur] = (p_matrix[sigma_n][n_cur] * p_matrix[beta_n][n_cur] -
                        p_matrix[tau_n][n_cur] * p_matrix[alpha_n][n_cur]) /
                        E_r / A_r /
                        (pow(p_matrix[alpha_n][n_cur], 2) + pow(p_matrix[beta_n][n_cur], 2));
                break;
            }
            case O_n: {
                p_matrix[p_num_cur][n_cur] = (p_matrix[kappa_n][n_cur] * ch(p_matrix[beta_n][n_cur] * x) +
                        p_matrix[delta_n][n_cur] * sh(p_matrix[beta_n][n_cur] * x)) *
                        Math.sin(p_matrix[alpha_n][n_cur] * x) +
                        (p_matrix[mu_n][n_cur] * sh(p_matrix[beta_n][n_cur] * x) +
                                p_matrix[nu_n][n_cur] * ch(p_matrix[beta_n][n_cur] * x)) *
                                Math.cos(p_matrix[alpha_n][n_cur] * x);
                break;
            }
            case P_n: {
                p_matrix[p_num_cur][n_cur] = (p_matrix[kappa_n][n_cur] * sh(p_matrix[beta_n][n_cur] * x) +
                        p_matrix[delta_n][n_cur] * ch(p_matrix[beta_n][n_cur] * x)) *
                        Math.cos(p_matrix[alpha_n][n_cur] * x) -
                        (p_matrix[mu_n][n_cur] * ch(p_matrix[beta_n][n_cur] * x) +
                                p_matrix[nu_n][n_cur] * sh(p_matrix[beta_n][n_cur] * x)) *
                                Math.sin(p_matrix[alpha_n][n_cur] * x);
                break;
            }
            case O_n_2: {
                p_matrix[p_num_cur][n_cur] = (p_matrix[tau_n][n_cur] / E_r / A_r * sh(p_matrix[beta_n][n_cur] * x) +
                        (p_matrix[delta_n][n_cur] * p_matrix[beta_n][n_cur] -
                                p_matrix[nu_n][n_cur] * p_matrix[alpha_n][n_cur]) *
                                ch(p_matrix[beta_n][n_cur] * x)) *
                        Math.sin(p_matrix[alpha_n][n_cur] * x) +
                        (p_matrix[sigma_n][n_cur] / E_r / A_r *
                                ch(p_matrix[beta_n][n_cur] * x) +
                                (p_matrix[nu_n][n_cur] * p_matrix[beta_n][n_cur] +
                                        p_matrix[delta_n][n_cur] * p_matrix[alpha_n][n_cur]) *
                                        sh(p_matrix[beta_n][n_cur] * x)) *
                                Math.cos(p_matrix[alpha_n][n_cur] * x);
                break;
            }
            case P_n_2: {
                p_matrix[p_num_cur][n_cur] = (p_matrix[tau_n][n_cur] / (E_r * A_r) * ch(p_matrix[beta_n][n_cur] * x) +
                        (p_matrix[delta_n][n_cur] * p_matrix[beta_n][n_cur] -
                                p_matrix[nu_n][n_cur] * p_matrix[alpha_n][n_cur]) *
                                sh(p_matrix[beta_n][n_cur] * x)) *
                        Math.cos(p_matrix[alpha_n][n_cur] * x) -
                        (p_matrix[sigma_n][n_cur] / (E_r * A_r) *
                                sh(p_matrix[beta_n][n_cur] * x) +
                                (p_matrix[nu_n][n_cur] * p_matrix[beta_n][n_cur] +
                                        p_matrix[delta_n][n_cur] * p_matrix[alpha_n][n_cur]) *
                                        ch(p_matrix[beta_n][n_cur] * x)) *
                                Math.sin(p_matrix[alpha_n][n_cur] * x);
                break;
            }
        }
    }

    private void c_p_matrix_first(int K, double E_r, double A_r,
                                  double omega,
                                  double dt, double a,
                                  double c, double x,
                                  int p_num_cur, int n_cur, Double[][] sgt) {
        switch (p_num_cur) {
            case alpha_n:
                p_matrix[p_num_cur][n_cur] =
                        (n_cur + 1) * omega / Math.sqrt(2) / a *
                                Math.sqrt(1 + Math.sqrt(1 + pow(c / (n_cur + 1) / omega, 2)));
                break;
            case beta_n:
                p_matrix[p_num_cur][n_cur] =
                        (n_cur + 1) * omega / Math.sqrt(2) / a *
                                Math.sqrt(-1 + Math.sqrt(1 + pow(c / (n_cur + 1) / omega, 2)));
                break;
            case sigma_n: {
                double temp = 0;
                for (int i = 0; i < K; ++i)
                    temp += (sgt[i][1] * 1000) * Math.cos(n_cur * omega * i * dt) * dt;
                p_matrix[p_num_cur][n_cur] = temp * omega / Math.PI;
                break;
            }
            case tau_n: {
                double temp = 0;
                for (int i = 0; i < K; ++i)
                    temp += (sgt[i][1] * 1000) * Math.sin(n_cur * omega * i * dt) * dt;
                p_matrix[p_num_cur][n_cur] = temp * omega / Math.PI;
                break;
            }
            case nu_n: {
                double temp = 0;
                for (int i = 0; i < K; ++i) temp += sgt[i][0] * Math.cos(n_cur * omega * i * dt) * dt;
                p_matrix[p_num_cur][n_cur] = temp * omega / Math.PI;
                break;
            }
            case delta_n: {
                double temp = 0;
                for (int i = 0; i < K; ++i) temp += sgt[i][0] * Math.sin(n_cur * omega * i * dt) * dt;
                p_matrix[p_num_cur][n_cur] = temp * omega / Math.PI;
                break;
            }
            default:
                c_p_matrix_common(E_r, A_r, x, p_num_cur, n_cur);
        }
    }

    private void c_p_matrix_next(double E_r_pre, double A_r_pre,
                                 double E_r, double A_r,
                                 double x, double L_pre, int p_num_cur, int n_cur) {
        switch (p_num_cur) {
            case alpha_n:
                break;
            case beta_n:
                break;
            case sigma_n: {
                if (n_cur == 0) return;
                p_matrix[p_num_cur][n_cur] = E_r_pre * A_r_pre * p_matrix[O_n_2][n_cur];
                break;
            }
            case tau_n: {
                p_matrix[p_num_cur][n_cur] = E_r_pre * A_r_pre * p_matrix[P_n_2][n_cur];
                break;
            }
            case nu_n: {
                if (n_cur > 0)
                    p_matrix[p_num_cur][n_cur] = p_matrix[O_n][n_cur];
                else
                    p_matrix[p_num_cur][n_cur] = p_matrix[sigma_n][0] / E_r_pre / A_r_pre * L_pre + p_matrix[nu_n][0];
                break;
            }
            case delta_n: {
                p_matrix[p_num_cur][n_cur] = p_matrix[P_n][n_cur];
                break;
            }
            default:
                c_p_matrix_common(E_r, A_r, x, p_num_cur, n_cur);
        }
    }


    static class rod_info {
        //直径 m
        double D_r;
        //长度/深度 m
        double L_r;
        //横截面积 平方米
        double A_r;
    }


    static class rods_group {
        rod_info[] rods;
        int rod_level;
        double A_r_avg;
        double D_r_avg;

        public rods_group(int level) {
            this.rod_level = level;
            rods = new rod_info[level];
        }
    }

    private rods_group group;

    private boolean handle_rods_data(Double[][] rods, double L_p) {
        if (rods == null || rods.length == 0) return false;
        group = new rods_group(rods.length);
        double temp = 0, temp2 = 0;
        for (int i = 0; i < rods.length; ++i) {
            rod_info info = new rod_info();
            info.D_r = rods[i][0] / 1000.0;//mm转m
            info.L_r = L_p * rods[i][1] / 100;//换算本级杆长度 m
            info.A_r = Math.PI * info.D_r * info.D_r / 4;//加入杆柱横截面积
            temp += info.A_r * info.L_r;
            temp2 += info.D_r * info.L_r;
            group.rods[i] = info;
        }
        group.A_r_avg = temp / L_p;//加入一个平均横截面积
        group.D_r_avg = temp2 / L_p;//加入一个平均杆直径
        return true;
    }

    private Double[][] c_multi_rod_diagram(Double[][] pc, double s, double f, double E_r, double rho_r,
                                           double mu_o, double L_p, double D_t, int n,
                                           Double[][] rods) {
        //对杆柱组合数据加工一下
        handle_rods_data(rods, L_p);

        int K = pc.length;
        //    rho_r = 7850;
        //    E_r = 2.1 * Math.pow(10, 11);
        //    D_r = 0.0545;
        double dt = 60 / f / K;
        double omega = 2 * Math.PI * f / 60;
        double a = Math.sqrt(E_r / rho_r);
        //    a = 4968;
        double c = c_C(group.A_r_avg, mu_o, D_t, group.D_r_avg, rho_r, omega, L_p, a);

        init_matrix(n);

        //计算傅立叶参数
        double L_r_pre = 0, A_r_pre = 0, x = 0;
        for (int level = 0; level < group.rod_level; ++level) {
            x += group.rods[level].L_r;
            double A_r = group.rods[level].A_r;
            double L_r = group.rods[level].L_r;
            if (level == 0) {
                for (int i = 0; i < n; ++i)
                    for (int j = 0; j < p_num; ++j)
                        c_p_matrix_first(K, E_r, A_r, omega, dt, a, c, -x, j, i, pc);
            } else {
                for (int i = 0; i < n; ++i)
                    for (int j = 0; j < p_num; ++j)
                        c_p_matrix_next(E_r, A_r_pre, E_r, A_r, -x, L_r_pre, j, i);
            }

            A_r_pre = A_r;
            L_r_pre = L_r;
        }

        //生成功图
        Double[][] duf = new Double[K][];
        for (int i = 0; i < K; ++i) {
            double t = i * dt;
            double tempA = 0;
            double tempP = 0;
            for (int j = 1; j < n; ++j) {
                tempA += p_matrix[O_n][j] * Math.cos(j * omega * t) + p_matrix[P_n][j] * Math.sin(j * omega * t);
                tempP += p_matrix[O_n_2][j] * Math.cos(j * omega * t) + p_matrix[P_n_2][j] * Math.sin(j * omega * t);
            }
            Double[] point = new Double[2];
            point[0] = (p_matrix[sigma_n][0] / 2 / E_r / A_r_pre * x + p_matrix[nu_n][0] / 2 + tempA);
            point[1] = (E_r * A_r_pre * (p_matrix[sigma_n][0] / 2 / E_r / A_r_pre + tempP));
            point[1] = point[1] / 1000.0;
            duf[i] = (point);
        }
        return duf;
    }

    private static final double ROD_E = 205800000000d;
    private static final double ROD_DENSITY = 8338d;
    private static final double MU_O = 0.032;
    private static final double D_T = 0.062;
    private static final int FOURIER_SERIES = 10;

    public Double[][] c_multi_rod_diagram(Double[][] pc, double s, double f, double L_p, Double[][] rods) {
        return c_multi_rod_diagram(pc, s, f, ROD_E, ROD_DENSITY, MU_O, L_p, D_T, FOURIER_SERIES, rods);
    }

    //@region test
    private static final PumpCardAlgo algo = new PumpCardAlgo();

    public static void main(String[] args) {
        Logger log = LoggerFactory.getLogger("pump-card-test");
        log.info("构建测试数据");
        Double[][] gt = JSONObject.parseObject("[[0,30.42],[0.01,30.98],[0.01,31.69],[0.02,32.4],[0.03,33.06],[0.04,33.69],[0.05,34.27],[0.06,34.79],[0.08,35.33],[0.09,35.82],[0.11,36.24],[0.13,36.58],[0.15,36.83],[0.17,37.13],[0.19,37.6],[0.21,38.23],[0.23,38.99],[0.26,39.86],[0.28,40.64],[0.31,41.26],[0.34,41.77],[0.36,42.22],[0.39,42.71],[0.42,43.42],[0.45,44.31],[0.48,45.33],[0.51,46.39],[0.54,47.32],[0.57,47.93],[0.6,48.27],[0.63,48.3],[0.66,48.05],[0.69,47.55],[0.73,46.91],[0.76,46.17],[0.79,45.46],[0.82,44.87],[0.86,44.52],[0.89,44.42],[0.92,44.51],[0.95,44.8],[0.99,45.2],[1.02,45.71],[1.05,46.27],[1.08,46.8],[1.11,47.19],[1.14,47.43],[1.18,47.47],[1.21,47.29],[1.24,46.99],[1.27,46.63],[1.31,46.25],[1.34,45.85],[1.37,45.52],[1.4,45.29],[1.43,45.18],[1.46,45.2],[1.49,45.37],[1.51,45.61],[1.54,45.9],[1.57,46.22],[1.59,46.52],[1.62,46.78],[1.65,46.94],[1.67,47],[1.7,46.93],[1.72,46.75],[1.75,46.49],[1.77,46.21],[1.79,45.91],[1.81,45.64],[1.84,45.45],[1.86,45.35],[1.88,45.36],[1.9,45.47],[1.92,45.66],[1.94,45.92],[1.96,46.25],[1.98,46.59],[1.99,46.88],[2.01,47.11],[2.03,47.23],[2.04,47.21],[2.06,47.06],[2.07,46.82],[2.09,46.49],[2.1,46.12],[2.11,45.77],[2.13,45.49],[2.14,45.32],[2.15,45.31],[2.15,45.47],[2.16,45.83],[2.17,46.23],[2.17,46.59],[2.18,46.85],[2.18,47.02],[2.18,47.07],[2.18,47.12],[2.19,47.2],[2.19,47.28],[2.19,47.25],[2.18,47.12],[2.18,46.86],[2.18,46.44],[2.17,45.87],[2.16,45.24],[2.15,44.46],[2.14,43.71],[2.13,42.97],[2.12,42.29],[2.11,41.59],[2.1,41.01],[2.08,40.44],[2.07,39.93],[2.06,39.48],[2.04,39.1],[2.02,38.71],[2,38.26],[1.98,37.78],[1.96,37.41],[1.94,37.25],[1.92,37.25],[1.89,37.47],[1.86,37.83],[1.84,38.16],[1.81,38.46],[1.78,38.74],[1.75,38.97],[1.72,39.07],[1.7,39.07],[1.67,38.94],[1.64,38.66],[1.61,38.32],[1.58,38],[1.55,37.75],[1.51,37.56],[1.48,37.49],[1.45,37.48],[1.42,37.57],[1.38,37.75],[1.35,38.01],[1.32,38.25],[1.28,38.45],[1.25,38.57],[1.21,38.6],[1.18,38.56],[1.14,38.45],[1.11,38.3],[1.08,38.11],[1.04,37.89],[1.01,37.66],[0.97,37.45],[0.94,37.32],[0.91,37.23],[0.87,37.22],[0.84,37.27],[0.81,37.33],[0.77,37.38],[0.74,37.4],[0.7,37.26],[0.67,36.72],[0.64,35.74],[0.61,34.51],[0.58,33.16],[0.55,31.91],[0.52,31.07],[0.49,30.7],[0.47,30.58],[0.44,30.42],[0.41,29.99],[0.38,29.33],[0.36,28.69],[0.33,28.2],[0.31,28.02],[0.28,28.14],[0.26,28.44],[0.24,28.68],[0.21,28.86],[0.19,29.12],[0.17,29.41],[0.15,29.7],[0.13,29.93],[0.11,30.03],[0.09,29.85],[0.08,29.64],[0.07,29.47],[0.05,29.3],[0.04,29.12],[0.03,28.97],[0.02,28.83],[0.01,28.76],[0.01,28.84],[0,29.05],[0,29.3],[0,29.51],[0,29.64],[0,29.68],[0,29.77],[0,30.02]]", Double[][].class);
        double s = 2.49;
        double n = 4.47;
        double L_p = 1400;
        log.info("original:==>\n{}\n\n\n", JSONObject.toJSONString(gt));

        Double[][] rod_module = {
                {19d, 75d},
                {22d, 20d},
                {25d, 5d},
        };

        long time = System.currentTimeMillis();
        Double[][] pc = algo.c_multi_rod_diagram(gt, s, n, L_p, rod_module);
        log.info("algo:==>\n{} \ntake {} ms\n\n", JSONObject.toJSONString(pc), System.currentTimeMillis() - time);
    }

    //endregion ---------------------------------------------------------------------------/
}
